@@ -1,6 +1,7 @@
 package MogileFS::DevFID;
 use strict;
 use warnings;
+use overload '""' => \&as_string;
 
 sub new {
     my ($class, $devarg, $fidarg) = @_;
@@ -17,6 +18,10 @@ sub new {
 sub devid { $_[0]{devid} }
 sub fidid { $_[0]{fidid} }
 
+sub as_string {
+    "DevFID[d=" . $_[0]{devid} . ";f=" . $_[0]{fidid} . "]";
+}
+
 sub device {
     my $self = shift;
     return $self->{dev} ||= MogileFS::Device->of_devid($self->{devid});
@@ -27,6 +32,13 @@ sub fid {
     return $self->{fid} ||= MogileFS::FID->new($self->{fidid});
 }
 
+# returns true if DevFID actually exists in database
+sub exists {
+    my $self = shift;
+    my $fid  = $self->fid;
+    return (grep { $_ == $self->{devid} } $fid->devids) ? 1 : 0;
+}
+
 sub url {
     my $self = shift;
     return $self->_make_full_url(0);
@@ -103,6 +115,22 @@ sub add_to_db {
     }
 }
 
+# Destroy a particular replica of a file via HTTP, remove it
+# from the tracker, and update the replication counts to be
+# accurate again.
+sub destroy {
+    my $self = shift;
+    my $httpfile = MogileFS::HTTPFile->at($self->url)
+        or die "Creation of HTTPFile object failed.";
+
+    $httpfile->delete
+        or die "Deletion of file via HTTP failed.";
+
+    my $sto = Mgd::get_store();
+    $sto->remove_fidid_from_devid($self->fidid, $self->devid);
+    $sto->update_devcount($self->fidid);
+}
+
 1;
 
 __END__
