@@ -25,7 +25,9 @@ sub work {
         # get db and note we're starting a run
         debug("Reaper running; looking for dead devices");
 
-        foreach my $dev (MogileFS::Device->dead_devices) {
+        foreach my $dev (grep { $_->dstate->is_perm_dead }
+                         MogileFS::Device->devices)
+        {
             my $devid = $dev->id;
             next if $all_empty{$devid};
 
